#include <Arduino.h>
#include "Display.h"


#define LV_BUF_SIZE     (ESP_PANEL_LCD_H_RES * ESP_PANEL_LCD_V_RES * 20 / 100)      // Local fraw buffer size (>= horizontal res. * 10)

ESP_Panel *panel;                                       // The display panel (LCD + touch)
lv_display_t *disp;                                     // LVGL display control structure
lv_indev_t *indev;                                      // LVGL input device control structure
void *drawBuffer1;                                      // LVGL draw buffer

TaskHandle_t lvgl_task_handle = nullptr;                // LVGL task handle
lv_mutex_t lvgl_mutex;                                  // LVGL mutex to protect LVGL against concurrent access


/**
 * Flushes the display with the provided pixel map within the specified area.
 *
 * @param disp Pointer to the display object.
 * @param area Pointer to the area to be updated on the display.
 * @param pixMap Pointer to the pixel map containing the updated pixel data.
 */
void DisplayFlush(lv_disp_t *disp, const lv_area_t *area, uint8_t *pixMap)
{
    panel->getLcd()->drawBitmap(area->x1, area->y1, area->x2 + 1, area->y2 + 1, pixMap);
    lv_disp_flush_ready(disp);
}

/**
 * Reads touch data from the LCD touch panel and updates the input device data.
 *
 * @param indev Pointer to the input device object.
 * @param data Pointer to the input device data structure to be updated.
 */
void TouchRead(lv_indev_t *indev, lv_indev_data_t *data)
{
    panel->getLcdTouch()->readData();
    bool touched = panel->getLcdTouch()->getTouchState();
    if(touched) {
        TouchPoint point = panel->getLcdTouch()->getPoint();
        data->point.x = point.x;
        data->point.y = point.y;
        data->state = LV_INDEV_STATE_PR;
    } else {
        data->state = LV_INDEV_STATE_REL;
    }
}

/**
 * @brief Function to increment the tick for LVGL.
 *
 * This function is used to increment the tick for LVGL by the specified period.
 *
 * @param arg Pointer to additional arguments (not used in this function).
 */
static void tick_increment(void *arg)
{
    /* Tell LVGL how many milliseconds have elapsed */
    lv_tick_inc(LVGL_PORT_TICK_PERIOD_MS);
}

/**
 * @brief Initializes the tick interface for LVGL.
 *
 * This function initializes the tick interface for LVGL by using esp_timer to generate a 2ms periodic event.
 *
 * @return esp_err_t Returns ESP_OK if the tick interface is successfully initialized, otherwise returns an error code.
 */
static esp_err_t tick_init(void)
{
    // Tick interface for LVGL (using esp_timer to generate 2ms periodic event)
    const esp_timer_create_args_t lvgl_tick_timer_args = {
        .callback = &tick_increment,
        .name = "LVGL tick"
    };
    esp_timer_handle_t lvgl_tick_timer = NULL;
    ESP_ERROR_CHECK(esp_timer_create(&lvgl_tick_timer_args, &lvgl_tick_timer));
    return esp_timer_start_periodic(lvgl_tick_timer, LVGL_PORT_TICK_PERIOD_MS * 1000);
}

void SetupDisplayOnLVGLThread()
{    
    // Allocate the draw buffers
    drawBuffer1 = heap_caps_malloc(LV_BUF_SIZE, MALLOC_CAP_DMA | MALLOC_CAP_INTERNAL);

    // Setup LVGL
    lv_init();
    disp = lv_display_create(ESP_PANEL_LCD_H_RES, ESP_PANEL_LCD_V_RES);
    lv_display_set_buffers(disp, drawBuffer1, NULL, LV_BUF_SIZE, LV_DISP_RENDER_MODE_PARTIAL);
    lv_display_set_flush_cb(disp, DisplayFlush);

    // Setup the input device
    indev = lv_indev_create();
    lv_indev_set_type(indev, LV_INDEV_TYPE_POINTER);
    lv_indev_set_read_cb(indev, TouchRead);

    lv_mutex_init(&lvgl_mutex);                         // Make lvgl thread safe

    // Setup the LVGL tick timer
    tick_init();                                        // Either use the LVGL tick timer or use the tick callback
    //lv_tick_set_cb(xTaskGetTickCount);                  // Set the tick callback for LVGL (no need to call lv_tick_inc)
}

/**
 * @brief Function to process the display using LVGL library.
 *
 * This function is responsible for initializing and handling the LVGL task.
 * It initializes LVGL such that this library is intizliased on the correct LVGL task
 * It locks the LVGL mutex, calls the LVGL timer handler,
 * and unlocks the mutex. It also adjusts the task delay based on the timer handler's return value.
 * The task delay is then used to delay the task execution using vTaskDelay function.
 *
 * @param arg Pointer to the argument (not used in this function).
 */
void ProcessDisplay(void *arg)
{
    ESP_LOGD(TAG, "Starting LVGL task");

    SetupDisplayOnLVGLThread();

    uint32_t task_delay_ms = LVGL_PORT_TASK_MAX_DELAY_MS;    
    while (1) {
        task_delay_ms = lv_timer_handler(); 
        if (task_delay_ms > LVGL_PORT_TASK_MAX_DELAY_MS) {
            task_delay_ms = LVGL_PORT_TASK_MAX_DELAY_MS;
        } else if (task_delay_ms < LVGL_PORT_TASK_MIN_DELAY_MS) {
            task_delay_ms = LVGL_PORT_TASK_MIN_DELAY_MS;
        }
        vTaskDelay(pdMS_TO_TICKS(task_delay_ms));
    }
}

/**
 * @brief Initializes and sets up the display panel and LVGL.
 * 
 * This function initializes the display panel.  It creates a display object, sets up the display buffer,
 * and creates the LVGL task which will do the deferred initialization of LVGL.
 * 
 * @note This function should be called before using any LVGL functions or rendering on the display.
 */
void SetupDisplay()
{
    // Initialize the display panel
    panel = new ESP_Panel();
    panel->init();
    panel->begin();

    // Setup the LVGL task which will iniitialize LVGL
    BaseType_t core_id = (LVGL_PORT_TASK_CORE < 0) ? tskNO_AFFINITY : LVGL_PORT_TASK_CORE;
    BaseType_t ret = xTaskCreatePinnedToCore(ProcessDisplay, "lvgl", LVGL_PORT_TASK_STACK_SIZE, NULL,
                                             LVGL_PORT_TASK_PRIORITY, &lvgl_task_handle, core_id);
}
